"""Fixtures for doctests."""

import doctest
import shutil
import urllib.request
from doctest import OutputChecker
from pathlib import Path

import duckdb
import pandas
import pytest
from pooch import retrieve

from quackosm.osm_extracts.extract import OsmExtractSource
from quackosm.osm_extracts.geofabrik import _get_geofabrik_index

IGNORE_RESULT = doctest.register_optionflag("IGNORE_RESULT")


class CustomOutputChecker(OutputChecker):
    """Custom doctest OutputChecker for ignoring logs from functions."""

    def check_output(self: doctest.OutputChecker, want: str, got: str, optionflags: int) -> bool:
        """Skips output checking if IGNORE_RESULT flag is present."""
        if IGNORE_RESULT & optionflags:
            return True
        return OutputChecker.check_output(self, want, got, optionflags)


doctest.OutputChecker = CustomOutputChecker  # type: ignore

LFS_DIRECTORY_URL = "https://github.com/kraina-ai/srai-test-files/raw/main/files/"

EXTRACTS_NAMES = ["monaco", "kiribati", "maldives"]


@pytest.fixture(autouse=True, scope="session")
def add_pbf_files(doctest_namespace):  # type: ignore
    """Download PBF files used in doctests."""
    download_directory = Path("files")
    download_directory.mkdir(parents=True, exist_ok=True)

    geofabrik_index = _get_geofabrik_index()
    for extract_name in EXTRACTS_NAMES:
        pbf_file_download_url = LFS_DIRECTORY_URL + f"{extract_name}-latest.osm.pbf"
        pbf_file_path = download_directory / f"{extract_name}.osm.pbf"
        geofabrik_download_path = geofabrik_index[geofabrik_index["name"] == extract_name].iloc[0][
            "file_name"
        ]
        geofabrik_pbf_file_path = download_directory / f"{geofabrik_download_path}.osm.pbf"
        urllib.request.urlretrieve(pbf_file_download_url, pbf_file_path)
        doctest_namespace[f"{extract_name}_pbf_path"] = pbf_file_path
        shutil.copy(pbf_file_path, geofabrik_pbf_file_path)



@pytest.fixture(autouse=True, scope="session")
def download_osm_extracts_indexes():  # type: ignore
    """Download OSM extract indexes files to cache."""
    download_directory = Path("cache")
    download_directory.mkdir(parents=True, exist_ok=True)

    for osm_extract in OsmExtractSource:
        if osm_extract == OsmExtractSource.any:
            continue

        file_name = f"{osm_extract.value.lower()}_index.geojson"
        file_download_url = LFS_DIRECTORY_URL + file_name

        retrieve(
            file_download_url,
            fname=file_name,
            path=download_directory,
            progressbar=True,
            known_hash=None,
        )


@pytest.fixture(autouse=True, scope="session")
def install_spatial_extension():  # type: ignore
    """Install duckdb spatial extension."""
    duckdb.install_extension("spatial")

@pytest.fixture(autouse=True, scope="session")  # type: ignore
def pandas_terminal_width() -> None:
    """Change pandas dataframe display options."""
    pandas.set_option("display.width", 90)
    pandas.set_option("display.max_colwidth", 35)



"""Global conftest for QuackOSM tests."""
import os

from pytest import Item


def pytest_runtest_setup(item: Item) -> None:
    """Setup python encoding before `pytest_runtest_call(item)`."""
    os.environ["PYTHONIOENCODING"] = "utf-8"



"""Tests for dedicated geocoding function."""

from typing import Union

import pytest
from osmnx.geocoder import geocode_to_gdf

from quackosm import geocode_to_geometry
from quackosm._exceptions import QueryNotGeocodedError


@pytest.mark.parametrize(  # type: ignore
    "query",
    [
        "Vatican",
        "Monaco",
        "Poland",
        ["United Kingdom", "Greater London"],
        ["Madrid", "Barcelona", "Seville"],
    ],
)
def test_geocoding(query: Union[str, list[str]]) -> None:
    """Test if geocoding works the same as osmnx."""
    assert geocode_to_gdf(query).unary_union.equals(geocode_to_geometry(query))


@pytest.mark.parametrize(  # type: ignore
    "query",
    [
        "Broadway",
        "nonexistent_query",
        ["Poland", "Broadway"],
    ],
)
def test_geocoding_errors(query: Union[str, list[str]]) -> None:
    """Test if geocoding throws error for two wrong queries."""
    with pytest.raises(QueryNotGeocodedError):
        geocode_to_geometry(query)



"""
OpenStreetMap.fr extracts.

This module contains wrapper for publically available OpenStreetMap.fr download server.
"""

import os
import re
from typing import Any, Optional

import geopandas as gpd
import requests
from tqdm import tqdm

from quackosm.osm_extracts._poly_parser import parse_polygon_file
from quackosm.osm_extracts.extract import (
    OpenStreetMapExtract,
    OsmExtractSource,
    extracts_to_geodataframe,
    load_index_decorator,
)

OPENSTREETMAP_FR_POLYGONS_INDEX_URL = "https://download.openstreetmap.fr/polygons"
OPENSTREETMAP_FR_EXTRACTS_INDEX_URL = "https://download.openstreetmap.fr/extracts"
OPENSTREETMAP_FR_INDEX_GDF: Optional[gpd.GeoDataFrame] = None

__all__ = ["_get_openstreetmap_fr_index"]


def _get_openstreetmap_fr_index() -> gpd.GeoDataFrame:
    global OPENSTREETMAP_FR_INDEX_GDF  # noqa: PLW0603

    if OPENSTREETMAP_FR_INDEX_GDF is None:
        OPENSTREETMAP_FR_INDEX_GDF = _load_openstreetmap_fr_index()

    return OPENSTREETMAP_FR_INDEX_GDF


@load_index_decorator(OsmExtractSource.osm_fr)
def _load_openstreetmap_fr_index() -> gpd.GeoDataFrame:  # pragma: no cover
    """
    Load available extracts from OpenStreetMap.fr download service.

    Returns:
        gpd.GeoDataFrame: Extracts index with metadata.
    """
    force_terminal = os.getenv("FORCE_TERMINAL_MODE", "false").lower() == "true"
    extracts = []
    with tqdm(disable=True if force_terminal else False) as pbar:
        extract_soup_objects = _gather_all_openstreetmap_fr_urls(
            OsmExtractSource.osm_fr.value, "/", pbar
        )
        pbar.set_description(OsmExtractSource.osm_fr.value)
        for soup_object, id_prefix, directory_url in extract_soup_objects:
            link = soup_object.find_parent("tr").find("a")
            name = link.text.replace("-latest.osm.pbf", "")
            polygon = parse_polygon_file(
                f"{OPENSTREETMAP_FR_POLYGONS_INDEX_URL}/{directory_url}{name}.poly"
            )
            if polygon is None:
                continue
            extracts.append(
                OpenStreetMapExtract(
                    id=f"{id_prefix}_{name}",
                    name=name,
                    parent=id_prefix,
                    url=f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}{link['href']}",
                    geometry=polygon,
                )
            )
            pbar.set_description_str(id_prefix)
            pbar.update()

    gdf = extracts_to_geodataframe(extracts)

    return gdf


def _gather_all_openstreetmap_fr_urls(
    id_prefix: str, directory_url: str, pbar: tqdm
) -> list[Any]:  # pragma: no cover
    """
    Iterate OpenStreetMap.fr extracts service page.

    Works recursively, by scraping whole available directory.

    Args:
        id_prefix (str): Prefix to be applies to extracts names.
        directory_url (str): Directory URL to load.
        pbar (tqdm): Progress bar.

    Returns:
        list[Any]: List of osm.fr extracts urls objects for further processing.
    """
    from bs4 import BeautifulSoup

    pbar.set_description_str(id_prefix)
    extract_soup_objects = []

    result = requests.get(
        f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}",
        headers={"User-Agent": "QuackOSM Python package (https://github.com/kraina-ai/quackosm)"},
    )
    soup = BeautifulSoup(result.text, "html.parser")

    extracts_urls = soup.find_all(string=re.compile("-latest\\.osm\\.pbf$"))
    pbar.total = (pbar.total or 0) + len(extracts_urls)
    pbar.refresh()
    extract_soup_objects.extend(
        [(extract_url, id_prefix, directory_url) for extract_url in extracts_urls]
    )

    directories = soup.find_all(src="/icons/folder.gif")
    for directory in directories:
        link = directory.find_parent("tr").find("a")
        name = link.text.replace("/", "")
        extract_soup_objects.extend(
            _gather_all_openstreetmap_fr_urls(
                id_prefix=f"{id_prefix}_{name}",
                directory_url=f"{directory_url}{link['href']}",
                pbar=pbar,
            )
        )

    return extract_soup_objects
